/********************************************

OpenGL program by Andreas Gustafsson (C) 2001

********************************************/

#define WIN_TIMER // To use Win-specific Queryperformance
#define BEZ_DIVS 8
#define NR_OBJS 512
#define PI 3.14159265

#include "stdio.h"
#include "math.h"
// This program requires the OpenGL and GLUT libraries
// You can obtain them for free from http://www.opengl.org
#include "GL/glut.h"
#include "GLTexture.h"

typedef struct vertexStruct
{
	float x,y,z;
} vertex;

double time=0,deltaTime=0;

#ifdef WIN_TIMER
static LARGE_INTEGER TimeFreq, TimeStart;
int updateFPS=0;
double curFPS=0;
char titleString[100];
#endif

CGLTexture box_front;
CGLTexture box_back;
CGLTexture box_left;
CGLTexture box_right;
CGLTexture box_up;
CGLTexture box_down;

CGLTexture featherTexture;

float rotY = 0.0;

vertex pos[NR_OBJS],rot[NR_OBJS],rotadd[NR_OBJS],color[NR_OBJS];
float size[NR_OBJS],bend[NR_OBJS];

float random(float high)
{
    return((float)((double)rand()/32768)*high);
}

void UpdateTimer()
{
	#ifdef WIN_TIMER
	static double oldTime;
   LARGE_INTEGER TimeNow;
   QueryPerformanceCounter(&TimeNow);
   time=((TimeNow.QuadPart-TimeStart.QuadPart)*1000/TimeFreq.QuadPart);
   time=time/1000;	
   deltaTime=time-oldTime;
   oldTime=time;

	if(deltaTime>0) curFPS+=1.f/deltaTime;
	updateFPS++;
	if(updateFPS==100)
	{
		sprintf(titleString,"OpenGL program by Andreas Gustafsson (Shadow/Noice) (C) 2001   -=[ FPS:%.0f ]=-",curFPS/100.0);
   	glutSetWindowTitle(titleString);
   	updateFPS=0;
   	curFPS=0;
   }
	#else
   deltaTime=0.01;
   time+=deltaTime;
	#endif
}

void Resize( GLsizei iWidth, GLsizei iHeight ) 
{ 
	glViewport (0,0,iWidth,iHeight); 
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(65.0,(GLfloat)iWidth/(GLfloat)iHeight, 0.1, 200.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void Idle()
{
	glutPostRedisplay();
}

void Init()
{
   box_front.LoadTGA("sky_front.tga");
	box_back.LoadTGA("sky_back.tga");
	box_left.LoadTGA("sky_left.tga");
	box_right.LoadTGA("sky_right.tga");
	box_up.LoadTGA("sky_up.tga");
	box_down.LoadTGA("sky_down.tga");
	featherTexture.LoadTGA("feather.tga");
	for(int i=0;i<NR_OBJS;i++)
	{
		float v1=random(PI);
		float v2=random(2*PI);
		float r=random(2)+random(40);
		pos[i].x=r*cos(v1);
		pos[i].y=r*cos(v2)*sin(v1);
		pos[i].z=r*sin(v2)*sin(v1);
		/*pos[i].x=20*(random(2)-1);
		pos[i].y=20*(random(2)-1);
		pos[i].z=20*(random(2)-1);*/
		rot[i].x=random(360);
		rot[i].y=random(360);
		rot[i].z=random(360);
		rotadd[i].x=random(150)-75;
		rotadd[i].y=random(150)-75;
		rotadd[i].z=random(150)-75;
		color[i].x=0.85+random(0.15);
		color[i].y=0.85+random(0.15);
		color[i].z=0.85+random(0.15);
		size[i]=0.8+random(0.4);
		bend[i]=0.25+random(0.5);
	}
}

void DrawSkybox()
{
   glEnable(GL_TEXTURE_2D);

   box_front.Use();
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glBegin(GL_QUADS);
	glTexCoord2f(1,0);
	glVertex3f(-1,-1,-1);
	glTexCoord2f(0,0);
	glVertex3f(1,-1,-1);
	glTexCoord2f(0,1);
	glVertex3f(1,1,-1);
	glTexCoord2f(1,1);
	glVertex3f(-1,1,-1);
	glEnd();

   box_up.Use();
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glBegin(GL_QUADS);
	glTexCoord2f(1,0);
	glVertex3f(-1,1,-1);
	glTexCoord2f(0,0);
	glVertex3f(1,1,-1);
	glTexCoord2f(0,1);
	glVertex3f(1,1,1);
	glTexCoord2f(1,1);
	glVertex3f(-1,1,1);
	glEnd();

   box_left.Use();
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glBegin(GL_QUADS);
	glTexCoord2f(1,1);
	glVertex3f(-1,1,1);
	glTexCoord2f(0,1);
	glVertex3f(-1,1,-1);
	glTexCoord2f(0,0);
	glVertex3f(-1,-1,-1);
	glTexCoord2f(1,0);
	glVertex3f(-1,-1,1);
	glEnd();

   box_right.Use();
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glBegin(GL_QUADS);
	glTexCoord2f(0,1);
	glVertex3f(1,1,1);
	glTexCoord2f(1,1);
	glVertex3f(1,1,-1);
	glTexCoord2f(1,0);
	glVertex3f(1,-1,-1);
	glTexCoord2f(0,0);
	glVertex3f(1,-1,1);
	glEnd();

   box_down.Use();
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glBegin(GL_QUADS);
	glTexCoord2f(1,1);
	glVertex3f(-1,-1,-1);
	glTexCoord2f(0,1);
	glVertex3f(1,-1,-1);
	glTexCoord2f(0,0);
	glVertex3f(1,-1,1);
	glTexCoord2f(1,0);
	glVertex3f(-1,-1,1);
	glEnd();

   box_back.Use();
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex3f(-1,-1,1);
	glTexCoord2f(1,0);
	glVertex3f(1,-1,1);
	glTexCoord2f(1,1);
	glVertex3f(1,1,1);
	glTexCoord2f(0,1);
	glVertex3f(-1,1,1);
	glEnd();
}

void Bezier(vertex p1,vertex p2,vertex p3,vertex p4,vertex p5,vertex p6)
{
	int   i,x,y;
   float u,bez02,bez12,bez22;
	vertex tmpV;

	u=0;
	glBegin(GL_TRIANGLE_STRIP);
	for(i=0;i<BEZ_DIVS;i++)
	{
   	bez02=(1-u)*(1-u);
      bez12=2*u*(1-u);
      bez22=u*u;
      tmpV.x=bez02*p1.x+bez12*p2.x+bez22*p3.x;
      tmpV.y=bez02*p1.y+bez12*p2.y+bez22*p3.y;
      tmpV.z=bez02*p1.z+bez12*p2.z+bez22*p3.z;
		glTexCoord2f(0,i/(float)(BEZ_DIVS-1));
      glVertex3f(tmpV.x,tmpV.y,tmpV.z);
      tmpV.x=bez02*p4.x+bez12*p5.x+bez22*p6.x;
      tmpV.y=bez02*p4.y+bez12*p5.y+bez22*p6.y;
      tmpV.z=bez02*p4.z+bez12*p5.z+bez22*p6.z;
		glTexCoord2f(1,i/(float)(BEZ_DIVS-1));
      glVertex3f(tmpV.x,tmpV.y,tmpV.z);
		u+=1.0/(float)BEZ_DIVS;;
	}
	glEnd();
}

void DrawScene() 
{ 	
	int i;
	vertex t1,t2,t3,t4,t5,t6,wind;
	float windStrength,tilt1,tilt2;
	
   UpdateTimer();
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ); 

  	rotY += deltaTime*15;
	tilt1=-25+50*sin(time*0.27);
	tilt2=45*sin(time*0.18);

   glPushMatrix(); 

   glPushMatrix();   
   glRotatef(tilt1, 1.0, 0.0, 0.0);
   glRotatef(rotY, 0.0, 1.0, 0.0);
   glRotatef(tilt2, 0.0, 0.0, 1.0);
   glColor3f(1.0, 1.0, 1.0);
   glDisable(GL_DEPTH_TEST);
   DrawSkybox();
   glEnable(GL_DEPTH_TEST);
   glPopMatrix();   
   
	glTranslatef (0.0, 0.0, 10.0*cos(time*0.2));
   glRotatef(tilt1, 1.0, 0.0, 0.0);
   glRotatef(rotY, 0.0, 1.0, 0.0);
   glRotatef(tilt2, 0.0, 0.0, 1.0);
	
	wind.x=1*sin(time);
	wind.y=0.25*cos(time*0.8);
	wind.z=1*sin(time*0.5)+0.5*cos(time*1.2);
	windStrength=1.65*(sin(time*0.6)+0.2*sin(time*2.3));
	
	featherTexture.Use();
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

	t1.x=0;			
	t1.y=-1;
	t1.z=0;

	t2.x=0;			
	t2.y=0;
	t2.z=0.5;

	t3.x=0;			
	t3.y=1.0;
	t3.z=0;

	t4.x=0.5;			
	t4.y=-1;
	t4.z=0;

	t5.x=0.5;			
	t5.y=0.0;
	t5.z=0.5;

	t6.x=0.5;			
	t6.y=1.0;
	t6.z=0;
		
	for(i=0;i<NR_OBJS;i++)
	{		
	  	glColor3f(color[i].x,color[i].y,color[i].z);
		glPushMatrix();		
		glTranslatef(pos[i].x,pos[i].y,pos[i].z);
   	glRotatef(rot[i].x, 1.0, 0.0, 0.0);
   	glRotatef(rot[i].y, 0.0, 1.0, 0.0);
   	glRotatef(rot[i].z, 0.0, 0.0, 1.0);		
   	glScalef(size[i],size[i],size[i]);
   	t2.z=bend[i];
   	t5.z=bend[i];
   	rot[i].x+=rotadd[i].x*deltaTime;
   	rot[i].y+=rotadd[i].y*deltaTime;
   	rot[i].z+=rotadd[i].z*deltaTime;
   	pos[i].x+=deltaTime*(sin(rot[i].x*PI/90)+windStrength*(1+sin(rot[i].x*PI/90))*wind.x);
   	pos[i].y+=deltaTime*(sin(rot[i].y*PI/90)+windStrength*(1+sin(rot[i].y*PI/90))*wind.y);
   	pos[i].z+=deltaTime*(sin(rot[i].z*PI/90)+windStrength*(1+sin(rot[i].z*PI/90))*wind.z);
		Bezier(t1,t2,t3,t4,t5,t6);
		glPopMatrix();
	}
	
	glDisable(GL_BLEND);	
	glPopMatrix();
   glutSwapBuffers(); 
}

void main(int argc, char **argv) 
{ 
   GLsizei iWidth = 640.0; 
   GLsizei iHeight = 480.0; 

   glutInit(&argc, argv);
   glutInitWindowPosition( 0, 0);
   glutInitWindowSize(iWidth, iHeight);
   glutInitDisplayMode( GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE );
   glutCreateWindow( "OpenGL Program by Andreas Gustafsson" );
   glutDisplayFunc(DrawScene );
   glutIdleFunc(Idle );
   glutReshapeFunc(Resize );
	Init();

   glClearColor( 0.0, 0.0, 0.0, 0.0 ); 
   glEnable(GL_DEPTH_TEST); 

	#ifdef WIN_TIMER
   QueryPerformanceFrequency(&TimeFreq);
   QueryPerformanceCounter(&TimeStart);
   #endif

   Resize(iWidth, iHeight); 
   glutMainLoop(); 

}
